"""App Configuration"""

# Django
from django.apps import AppConfig

# AA Example App
from taxsystem import __version__


class TaxSystemConfig(AppConfig):
    """App Config"""

    default_auto_field = "django.db.models.AutoField"
    author = "Geuthur"
    name = "taxsystem"
    label = "taxsystem"
    verbose_name = f"Tax System v{__version__}"

    def ready(self):
        # pylint: disable=import-outside-toplevel,unused-import
        import taxsystem.signals



from ninja import NinjaAPI
from ninja.security import django_auth

from django.conf import settings

from taxsystem.api import taxsystem
from taxsystem.hooks import get_extension_logger

logger = get_extension_logger(__name__)

api = NinjaAPI(
    title="Geuthur API",
    version="0.1.0",
    urls_namespace="taxsystem:new_api",
    auth=django_auth,
    csrf=True,
    openapi_url=settings.DEBUG and "/openapi.json" or "",
)

# Add the taxsystem endpoints
taxsystem.setup(api)



from .filters import FilterAmount, FilterBase, FilterReason, SmartFilter, SmartGroup
from .general import General
from .tax import (
    Members,
    OwnerAudit,
    Payments,
)
from .wallet import CorporationWalletDivision, CorporationWalletJournalEntry



"""Wallet Helpers"""

from django.utils import timezone
from eveuniverse.models import EveEntity

from taxsystem.errors import DatabaseError
from taxsystem.hooks import get_extension_logger
from taxsystem.models.tax import (
    OwnerAudit,
)
from taxsystem.models.wallet import (
    CorporationWalletDivision,
    CorporationWalletJournalEntry,
)
from taxsystem.providers import esi
from taxsystem.task_helpers.etag_helpers import NotModifiedError, etag_results
from taxsystem.task_helpers.general_helpers import get_corp_token

logger = get_extension_logger(__name__)


def update_corporation_wallet_division(corp_id, force_refresh=False):
    audit_corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)

    req_scopes = [
        "esi-wallet.read_corporation_wallets.v1",
        "esi-characters.read_corporation_roles.v1",
        "esi-corporations.read_divisions.v1",
    ]
    req_roles = ["CEO", "Director"]

    token = get_corp_token(corp_id, req_scopes, req_roles)
    names = {}

    if token:
        division_names = (
            esi.client.Corporation.get_corporations_corporation_id_divisions(
                corporation_id=audit_corp.corporation.corporation_id,
                token=token.valid_access_token(),
            ).result()
        )

        for division in division_names.get("wallet"):
            names[division.get("division")] = division.get("name")

    req_roles = ["CEO", "Director", "Accountant", "Junior_Accountant"]

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        logger.debug("No valid token for: %s", audit_corp.corporation.corporation_name)
        return "No Tokens"

    try:
        divisions_items_ob = esi.client.Wallet.get_corporations_corporation_id_wallets(
            corporation_id=audit_corp.corporation.corporation_id
        )

        division_items = etag_results(
            divisions_items_ob, token, force_refresh=force_refresh
        )

        for division in division_items:
            _division_item, _ = CorporationWalletDivision.objects.update_or_create(
                corporation=audit_corp,
                division=division.get("division"),
                defaults={
                    "balance": division.get("balance"),
                    "name": names.get(division.get("division"), "Unknown"),
                },
            )

            if _division_item:
                update_corp_wallet_journal(
                    corp_id, division.get("division"), force_refresh=force_refresh
                )  # inline not async

    except NotModifiedError:
        logger.debug(
            "No New wallet data for: %s",
            audit_corp.corporation.corporation_name,
        )

    audit_corp.last_update_wallet = timezone.now()
    audit_corp.save()

    return ("Finished wallet divs for: %s", audit_corp.corporation.corporation_name)


# pylint: disable=too-many-locals
def update_corp_wallet_journal(corp_id, wallet_division, force_refresh=False):
    audit_corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)

    division = CorporationWalletDivision.objects.get(
        corporation=audit_corp, division=wallet_division
    )

    logger.debug(
        "Updating wallet transactions for: %s (Div: %s)",
        audit_corp.corporation.corporation_name,
        division,
    )

    req_scopes = [
        "esi-wallet.read_corporation_wallets.v1",
        "esi-characters.read_corporation_roles.v1",
    ]

    req_roles = ["CEO", "Director", "Accountant", "Junior_Accountant"]

    token = get_corp_token(corp_id, req_scopes, req_roles)

    if not token:
        logger.debug("No valid token for: %s", audit_corp.corporation.corporation_name)
        return "No Tokens"

    try:
        _current_journal = set(
            list(
                CorporationWalletJournalEntry.objects.filter(division=division)
                .order_by("-date")
                .values_list("entry_id", flat=True)[:20000]
            )
        )
        _current_eve_ids = set(
            list(EveEntity.objects.all().values_list("id", flat=True))
        )

        current_page = 1
        total_pages = 1
        _new_names = []
        while current_page <= total_pages:
            journal_items_ob = esi.client.Wallet.get_corporations_corporation_id_wallets_division_journal(
                corporation_id=audit_corp.corporation.corporation_id,
                division=wallet_division,
                page=current_page,
            )
            journal_items = etag_results(
                journal_items_ob, token, force_refresh=force_refresh
            )

            _min_time = timezone.now()
            items = []
            for item in journal_items:
                _min_time = min(_min_time, item.get("date"))

                if item.get("id") not in _current_journal:
                    if item.get("second_party_id") not in _current_eve_ids:
                        _new_names.append(item.get("second_party_id"))
                        _current_eve_ids.add(item.get("second_party_id"))
                    if item.get("first_party_id") not in _current_eve_ids:
                        _new_names.append(item.get("first_party_id"))
                        _current_eve_ids.add(item.get("first_party_id"))

                    wallet_item = (
                        CorporationWalletJournalEntry(  # pylint: disable=duplicate-code
                            division=division,
                            amount=item.get("amount"),
                            balance=item.get("balance"),
                            context_id=item.get("context_id"),
                            context_id_type=item.get("context_id_type"),
                            date=item.get("date"),
                            description=item.get("description"),
                            first_party_id=item.get("first_party_id"),
                            entry_id=item.get("id"),
                            reason=item.get("reason"),
                            ref_type=item.get("ref_type"),
                            second_party_id=item.get("second_party_id"),
                            tax=item.get("tax"),
                            tax_receiver_id=item.get("tax_receiver_id"),
                        )
                    )

                    items.append(wallet_item)

            logger.debug(
                "Corp %s Div %s, Page: %s, New Transactions! %s, New Names %s",
                corp_id,
                wallet_division,
                current_page,
                len(items),
                _new_names,
            )

            # Create Entities
            EveEntity.objects.bulk_resolve_ids(_new_names)
            # Check if created
            all_exist = EveEntity.objects.filter(id__in=_new_names).count() == len(
                _new_names
            )

            if all_exist:
                CorporationWalletJournalEntry.objects.bulk_create(items)
            else:
                raise DatabaseError("DB Fail")

            current_page += 1
    except NotModifiedError:
        logger.debug(
            "No New wallet data for: Div: %s Corp: %s",
            audit_corp.corporation.corporation_name,
            wallet_division,
        )
    return ("Finished wallet journal for: %s", audit_corp.corporation.corporation_name)



# Django
from django.db import models

# AA Voices of War
from taxsystem.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class PaymentSystemQuerySet(models.QuerySet):
    pass


class PaymentSystemManagerBase(models.Manager):
    pass


PaymentSystemManager = PaymentSystemManagerBase.from_queryset(PaymentSystemQuerySet)


class PaymentsQuerySet(models.QuerySet):
    pass


class PaymentsManagerBase(models.Manager):
    pass


PaymentsManager = PaymentsManagerBase.from_queryset(PaymentsQuerySet)



"""
App Settings
"""

import sys

# Django
from app_utils.app_settings import clean_setting

IS_TESTING = sys.argv[1:2] == ["test"]

# EVE Online Swagger
EVE_BASE_URL = "https://esi.evetech.net/"
EVE_API_URL = "https://esi.evetech.net/latest/"
EVE_BASE_URL_REGEX = r"^http[s]?:\/\/esi.evetech\.net\/"

# Fuzzwork
FUZZ_BASE_URL = "https://www.fuzzwork.co.uk/"
FUZZ_API_URL = "https://www.fuzzwork.co.uk/api/"
FUZZ_BASE_URL_REGEX = r"^http[s]?:\/\/(www\.)?fuzzwork\.co\.uk\/"

# ZKillboard
ZKILLBOARD_BASE_URL = "https://zkillboard.com/"
ZKILLBOARD_API_URL = "https://zkillboard.com/api/"
ZKILLBOARD_BASE_URL_REGEX = r"^http[s]?:\/\/zkillboard\.com\/"
ZKILLBOARD_KILLMAIL_URL_REGEX = r"^http[s]?:\/\/zkillboard\.com\/kill\/\d+\/"

# Set Test Mode True or False

# Set Naming on Auth Hook
TAXSYSTEM_APP_NAME = clean_setting("TAXSYSTEM_APP_NAME", "Tax System")

# If True you need to set up the Logger
TAXSYSTEM_LOGGER_USE = clean_setting("TAXSYSTEM_LOGGER_USE", False)

# Skip Dates for Audit

# Member Skip Date in Days
TAXSYSTEM_CORP_MEMBERS_SKIP_DATE = clean_setting("TAXSYSTEM_CORP_MEMBERS_SKIP_DATE", 1)
# Wallet Skip Date in Hours
TAXSYSTEM_CORP_WALLET_SKIP_DATE = clean_setting("TAXSYSTEM_CORP_WALLET_SKIP_DATE", 1)
# Payment Skip Date in Hours
TAXSYSTEM_CORP_PAYMENTS_SKIP_DATE = clean_setting(
    "TAXSYSTEM_CORP_PAYMENTS_SKIP_DATE", 1
)
# Filter Skip Date in Hours
TAXSYSTEM_CORP_PAYMENT_SYSTEM_SKIP_DATE = clean_setting(
    "TAXSYSTEM_CORP_PAYMENT_SYSTEM_SKIP_DATE", 1
)



from ninja import NinjaAPI

from taxsystem.hooks import get_extension_logger
from taxsystem.models.tax import OwnerAudit

logger = get_extension_logger(__name__)


class AdminApiEndpoints:
    tags = ["Admin"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "corporation/admin/",
            response={200: list, 403: str},
            tags=self.tags,
        )
        def get_corporation_admin(request):
            corporations = OwnerAudit.objects.visible_to(request.user)

            if corporations is None:
                return 403, "Permission Denied"

            corporation_dict = {}

            for corporation in corporations:
                # pylint: disable=broad-exception-caught
                try:
                    corporation_dict[corporation.corporation.corporation_id] = {
                        "corporation_id": corporation.corporation.corporation_id,
                        "corporation_name": corporation.corporation.corporation_name,
                    }
                except Exception:
                    continue

            output = []
            output.append({"corporation": corporation_dict})

            return output



"""Models for Tax System."""

# Django
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.authentication.models import UserProfile
from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo

from taxsystem.managers.payment_manager import PaymentsManager, PaymentSystemManager
from taxsystem.managers.tax_manager import MembersManager, OwnerAuditManager


class OwnerAudit(models.Model):
    """Tax System Audit model for app"""

    objects = OwnerAuditManager()

    name = models.CharField(
        max_length=255,
    )

    corporation = models.OneToOneField(
        EveCorporationInfo, on_delete=models.CASCADE, related_name="+"
    )

    alliance = models.ForeignKey(
        EveAllianceInfo,
        on_delete=models.CASCADE,
        related_name="alliance",
        blank=True,
        null=True,
    )

    active = models.BooleanField(default=False)

    last_update_wallet = models.DateTimeField(null=True, blank=True)

    last_update_members = models.DateTimeField(null=True, blank=True)

    last_update_payments = models.DateTimeField(null=True, blank=True)

    last_update_payment_system = models.DateTimeField(null=True, blank=True)

    tax_amount = models.DecimalField(
        max_digits=16,
        decimal_places=0,
        help_text=_("Tax Amount in ISK that is set for the corporation. Max 16 Digits"),
        default=0,
        validators=[MaxValueValidator(9999999999999999)],
    )

    tax_period = models.PositiveIntegerField(
        help_text=_(
            "Tax Period in days for the corporation. Max 365 days. Default: 30 days"
        ),
        default=30,
        validators=[MaxValueValidator(365)],
    )

    def __str__(self):
        return f"{self.corporation.corporation_name} - Audit Data"

    @classmethod
    def get_esi_scopes(cls) -> list[str]:
        """Return list of required ESI scopes to fetch."""
        return [
            # General
            "esi-corporations.read_corporation_membership.v1",
            "esi-corporations.track_members.v1",
            "esi-characters.read_corporation_roles.v1",
            # wallets
            "esi-wallet.read_corporation_wallets.v1",
            "esi-corporations.read_divisions.v1",
        ]

    class Meta:
        default_permissions = ()
        verbose_name = _("Tax System Audit")
        verbose_name_plural = _("Tax System Audits")


class Members(models.Model):
    """Tax System Member model for app"""

    class States(models.TextChoices):
        ACTIVE = "active", _("Active")
        MISSING = "missing", _("Missing")
        NOACCOUNT = "noaccount", _("Unregistered")
        IS_ALT = "is_alt", _("Is Alt")

    character_name = models.CharField(max_length=100, db_index=True)

    character_id = models.PositiveIntegerField(primary_key=True)

    corporation = models.ForeignKey(
        OwnerAudit, on_delete=models.CASCADE, related_name="owner"
    )

    status = models.CharField(
        _("Status"), max_length=10, choices=States.choices, blank=True, default="active"
    )

    logon = models.DateTimeField(null=True, blank=True)

    logged_off = models.DateTimeField(null=True, blank=True)

    joined = models.DateTimeField(null=True, blank=True)

    notice = models.TextField(null=True, blank=True)

    class Meta:
        default_permissions = ()
        verbose_name = _("Tax Member System")
        verbose_name_plural = _("Tax Member Systems")

    def __str__(self):
        return f"{self.character_name} - {self.character_id}"

    objects = MembersManager()

    @property
    def is_active(self) -> bool:
        return self.status == self.States.ACTIVE

    @property
    def is_missing(self) -> bool:
        return self.status == self.States.MISSING

    @property
    def is_noaccount(self) -> bool:
        return self.status == self.States.NOACCOUNT

    @property
    def is_alt(self) -> bool:
        return self.status == self.States.IS_ALT

    @property
    def is_faulty(self) -> bool:
        return self.status in [self.States.MISSING, self.States.NOACCOUNT]


class PaymentSystem(models.Model):
    """Tax Payment System model for app"""

    class States(models.TextChoices):
        ACTIVE = "active", _("Active")
        INACTIVE = "inactive", _("Inactive")
        DEACTIVATED = "deactivated", _("Deactivated")

    name = models.CharField(
        max_length=100,
    )

    corporation = models.ForeignKey(
        OwnerAudit, on_delete=models.CASCADE, related_name="+"
    )

    user = models.OneToOneField(UserProfile, on_delete=models.CASCADE, related_name="+")

    date = models.DateTimeField(auto_now_add=True, null=True, blank=True)

    status = models.CharField(
        _("Status"), max_length=16, choices=States.choices, blank=True, default=""
    )

    payment_pool = models.DecimalField(
        max_digits=16,
        decimal_places=0,
        default=0,
        help_text=_(
            "Payment Pool in ISK that is set for the corporation. Max 16 Digits"
        ),
        validators=[
            MaxValueValidator(9999999999999999),
            MinValueValidator(-9999999999999999),
        ],
    )

    last_paid = models.DateTimeField(null=True, blank=True)

    notice = models.TextField(null=True, blank=True)

    class Meta:
        default_permissions = ()
        verbose_name = _("Tax Payment System")
        verbose_name_plural = _("Tax Payment Systems")

    def __str__(self):
        return f"{self.name} - {self.date} - {self.status}"

    def get_payment_status(self) -> str:
        return self.get_status_display()

    def get_alt_ids(self) -> list[int]:
        return list(
            self.user.user.character_ownerships.all().values_list(
                "character__character_id", flat=True
            )
        )

    @property
    def is_active(self) -> bool:
        return self.status == self.States.ACTIVE

    @property
    def is_inactive(self) -> bool:
        return self.status == self.States.INACTIVE

    @property
    def is_deactivated(self) -> bool:
        return self.status == self.States.DEACTIVATED

    @property
    def has_paid(self) -> bool:
        """Return True if user has paid the set amount or if last_paid is within the tax period."""
        if self.payment_pool >= self.corporation.tax_amount:
            return True
        if self.last_paid:
            return timezone.now() - self.last_paid < timezone.timedelta(
                days=self.corporation.tax_period
            )
        return False

    objects = PaymentSystemManager()


class Payments(models.Model):
    """Tax Payments model for app"""

    class States(models.TextChoices):
        PAID = "paid", _("Paid")
        PENDING = "pending", _("Pending")
        FAILED = "failed", _("Failed")
        NEEDS_APPROVAL = "needs_approval", _("Needs Approval")

    class Approval(models.TextChoices):
        APPROVED = "approved", _("Approved")
        PENDING = "pending", _("Pending")
        REJECTED = "rejected", _("Rejected")

    class Systems(models.TextChoices):
        AUTOMATIC = "automatic", _("Automatic")
        MANUAL = "manual", _("Manual")

    name = models.CharField(max_length=100)

    entry_id = models.BigIntegerField()

    payment_user = models.ForeignKey(
        PaymentSystem, on_delete=models.CASCADE, related_name="payment_user"
    )

    date = models.DateTimeField(default=timezone.now, null=True, blank=True)

    amount = models.DecimalField(max_digits=12, decimal_places=0)

    payment_status = models.CharField(
        _("Status"), max_length=16, choices=States.choices, blank=True, default=""
    )

    payment_date = models.DateTimeField(null=True, blank=True)

    reason = models.TextField(null=True, blank=True)

    approved = models.CharField(
        _("Pending"), max_length=16, choices=Approval.choices, blank=True, default=""
    )

    system = models.CharField(
        _("System"), max_length=16, choices=Systems.choices, blank=True, default=""
    )

    notified = models.BooleanField(default=False)

    class Meta:
        default_permissions = ()
        verbose_name = _("Tax Payments")
        verbose_name_plural = _("Tax Payments")

    @property
    def is_human(self) -> bool:
        return self.system == self.Systems.MANUAL

    @property
    def is_automatic(self) -> bool:
        return self.system == self.Systems.AUTOMATIC

    def __str__(self):
        return f"{self.payment_user.name} - {self.date} - {self.amount} - {self.payment_status}"

    def get_payment_status(self) -> str:
        return self.get_payment_status_display()

    def get_approval_status(self) -> str:
        return self.get_approved_display()

    def formatted_payment_date(self) -> str:
        if self.payment_date:
            return timezone.localtime(self.payment_date).strftime("%Y-%m-%d %H:%M:%S")
        return _("No payment date")

    objects = PaymentsManager()



"""Tax System for a TestView class."""

from http import HTTPStatus

from django.test import RequestFactory, TestCase
from django.urls import reverse

from app_utils.testdata_factories import UserMainFactory

from taxsystem.views import index


class TestViews(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.user = UserMainFactory(
            permissions=[
                "taxsystem.basic_access",
            ]
        )

    def test_view(self):
        request = self.factory.get(reverse("taxsystem:index"))
        request.user = self.user
        response = index(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)



"""Tax Helpers"""

from django.db import transaction
from django.utils import timezone

from taxsystem.hooks import get_extension_logger
from taxsystem.models.filters import SmartGroup
from taxsystem.models.tax import (
    OwnerAudit,
    Payments,
    PaymentSystem,
)
from taxsystem.models.wallet import CorporationWalletJournalEntry

logger = get_extension_logger(__name__)


# pylint: disable=too-many-locals
def update_corporation_payments(corp_id):
    """Update corporation members"""
    audit_corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)

    logger.debug(
        "Updating payments for: %s",
        audit_corp.corporation.corporation_name,
    )

    payment_users = PaymentSystem.objects.filter(corporation=audit_corp)

    if not payment_users:
        return ("No Payment Users for %s", audit_corp.corporation.corporation_name)

    users = {}

    for user in payment_users:
        alts = user.get_alt_ids()
        users[user] = alts

    journal = CorporationWalletJournalEntry.objects.filter(
        division__corporation=audit_corp, ref_type__in=["player_donation"]
    ).order_by("-date")

    _current_entry_ids = set(
        Payments.objects.filter(payment_user__corporation=audit_corp).values_list(
            "entry_id", flat=True
        )
    )

    items = []
    for entry in journal:
        # Skip if already processed
        if entry.entry_id in _current_entry_ids:
            continue
        for user, alts in users.items():
            if entry.first_party.id in alts:
                payment_item = Payments(
                    entry_id=entry.entry_id,
                    name=user.name,
                    payment_user=user,
                    date=timezone.now(),
                    amount=entry.amount,
                    payment_status=Payments.States.PENDING,
                    payment_date=entry.date,
                    reason=entry.reason,
                    approved=Payments.Approval.PENDING,
                )
                items.append(payment_item)

    Payments.objects.bulk_create(items, ignore_conflicts=True)

    audit_corp.last_update_payments = timezone.now()
    audit_corp.save()

    logger.debug(
        "Finished %s Payments for %s",
        len(items),
        audit_corp.corporation.corporation_name,
    )

    return ("Finished Payments for %s", audit_corp.corporation.corporation_name)


def update_corporation_payments_filter(corp_id, runs=0):
    """Update Filtered Payments for Corporation"""
    audit_corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)

    logger.debug(
        "Updating payment system for: %s",
        audit_corp.corporation.corporation_name,
    )

    payments = Payments.objects.filter(
        payment_user__corporation=audit_corp, payment_status=Payments.States.PENDING
    )

    _current_payment_ids = set(payments.values_list("id", flat=True))
    _automatic_payment_ids = []

    # Check for any automatic payments
    try:
        filters = SmartGroup.objects.get(corporation=audit_corp)
        if filters:
            payments = filters.filter(payments)
            for payment in payments:
                if payment.payment_status == Payments.States.PENDING:
                    # Ensure all transers are processed in a single transaction
                    with transaction.atomic():
                        payment.payment_status = Payments.States.PAID
                        payment.approved = Payments.Approval.APPROVED
                        payment.system = Payments.Systems.AUTOMATIC

                        # Update payment pool for user
                        PaymentSystem.objects.filter(
                            corporation=audit_corp, user=payment.payment_user.user
                        ).update(
                            payment_pool=payment.payment_user.payment_pool
                            + payment.amount
                        )

                        payment.save()
                        runs = runs + 1
                        _automatic_payment_ids.append(payment.pk)
    except SmartGroup.DoesNotExist:
        pass

    # Check for any payments that need approval
    needs_approval = _current_payment_ids - set(_automatic_payment_ids)
    approvals = Payments.objects.filter(
        id__in=needs_approval, payment_status=Payments.States.PENDING
    )

    for payment in approvals:
        payment.payment_status = Payments.States.NEEDS_APPROVAL
        payment.approved = Payments.Approval.PENDING
        payment.save()
        runs = runs + 1

    # Check approved payments
    approved = Payments.objects.filter(
        payment_user__corporation=audit_corp,
        payment_status=Payments.States.NEEDS_APPROVAL,
        approved=Payments.Approval.APPROVED,
    )

    for payment in approved:
        with transaction.atomic():
            payment.payment_status = Payments.States.PAID
            PaymentSystem.objects.filter(
                corporation=audit_corp, user=payment.payment_user.user
            ).update(payment_pool=payment.payment_user.payment_pool + payment.amount)
            payment.save()
            runs = runs + 1

    # Check Payment Period
    payday = PaymentSystem.objects.filter(
        corporation=audit_corp, status=PaymentSystem.States.ACTIVE
    )

    for user in payday:
        if user.last_paid is None:
            # First Period is free
            user.last_paid = timezone.now()
        if timezone.now() - user.last_paid >= timezone.timedelta(
            days=audit_corp.tax_period
        ):
            user.payment_pool -= audit_corp.tax_amount
            user.last_paid = timezone.now()
            runs = runs + 1
        user.save()

    audit_corp.last_update_payment_system = timezone.now()
    audit_corp.save()

    logger.debug(
        "Finished %s: Payment System entrys for %s",
        runs,
        audit_corp.corporation.corporation_name,
    )

    return ("Finished Payment System for %s", audit_corp.corporation.corporation_name)



# Django
from django.db import models

# AA Voices of War
from taxsystem.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class OwnerAuditQuerySet(models.QuerySet):
    def visible_to(self, user):
        # superusers get all visible
        if user.is_superuser:
            logger.debug(
                "Returning all corps for superuser %s.",
                user,
            )
            return self

        if user.has_perm("taxsystem.manage_corps"):
            logger.debug("Returning all corps for Tax Audit Manager %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            query = None

            if user.has_perm("taxsystem.manage_own_corp"):
                query = models.Q(corporation__corporation_id=char.corporation_id)

            logger.debug("Returning own corps for User %s.", user)

            if query is None:
                return self.none()

            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()


class OwnerAuditManagerBase(models.Manager):
    def visible_to(self, user):
        return self.get_queryset().visible_to(user)


OwnerAuditManager = OwnerAuditManagerBase.from_queryset(OwnerAuditQuerySet)


class MembersQuerySet(models.QuerySet):
    pass


class MembersManagerBase(models.Manager):
    pass


MembersManager = MembersManagerBase.from_queryset(MembersQuerySet)



# Generated by Django 4.2.11 on 2025-02-08 12:47

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
        ("authentication", "0024_alter_userprofile_language"),
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app"),
                    ("manage_access", "Can manage Tax System"),
                    ("create_access", "Can add Corporation"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="FilterAmount",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("description", models.CharField(max_length=255)),
                ("amount", models.DecimalField(decimal_places=0, max_digits=12)),
            ],
            options={
                "verbose_name": "Filter Amount",
                "verbose_name_plural": "Filter Amounts",
            },
        ),
        migrations.CreateModel(
            name="FilterReason",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("description", models.CharField(max_length=255)),
                ("reason", models.CharField(max_length=255)),
            ],
            options={
                "verbose_name": "Filter Reason",
                "verbose_name_plural": "Filter Reasons",
            },
        ),
        migrations.CreateModel(
            name="OwnerAudit",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("active", models.BooleanField(default=False)),
                ("last_update_wallet", models.DateTimeField(blank=True, null=True)),
                ("last_update_members", models.DateTimeField(blank=True, null=True)),
                ("last_update_payments", models.DateTimeField(blank=True, null=True)),
                ("last_update_filters", models.DateTimeField(blank=True, null=True)),
                (
                    "alliance",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="alliance",
                        to="eveonline.eveallianceinfo",
                    ),
                ),
                (
                    "corporation",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveonline.evecorporationinfo",
                    ),
                ),
            ],
            options={
                "verbose_name": "Tax System Audit",
                "verbose_name_plural": "Tax System Audits",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="SmartFilter",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.PositiveIntegerField()),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Smart Filter Binding",
                "verbose_name_plural": "Smart Filters Catalog",
            },
        ),
        migrations.CreateModel(
            name="SmartGroup",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("description", models.CharField(max_length=255)),
                ("name", models.CharField(max_length=100)),
                ("last_update", models.DateTimeField(auto_now=True)),
                ("enabled", models.BooleanField(default=True)),
                (
                    "corporation",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="filter_sets",
                        to="taxsystem.owneraudit",
                    ),
                ),
                ("filters", models.ManyToManyField(to="taxsystem.smartfilter")),
            ],
        ),
        migrations.CreateModel(
            name="PaymentSystem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("date", models.DateTimeField(auto_now_add=True, null=True)),
                (
                    "status",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("active", "Active"),
                            ("inactive", "Inactive"),
                            ("deactivated", "Deactivated"),
                        ],
                        default="",
                        max_length=16,
                        verbose_name="Status",
                    ),
                ),
                ("notice", models.TextField(blank=True, null=True)),
                (
                    "corporation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="taxsystem.owneraudit",
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="authentication.userprofile",
                    ),
                ),
            ],
            options={
                "verbose_name": "Tax Payment System",
                "verbose_name_plural": "Tax Payment Systems",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Payments",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("entry_id", models.BigIntegerField()),
                (
                    "date",
                    models.DateTimeField(
                        blank=True, default=django.utils.timezone.now, null=True
                    ),
                ),
                ("amount", models.DecimalField(decimal_places=0, max_digits=12)),
                (
                    "payment_status",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("paid", "Paid"),
                            ("pending", "Pending"),
                            ("failed", "Failed"),
                        ],
                        default="",
                        max_length=10,
                        verbose_name="Status",
                    ),
                ),
                ("payment_date", models.DateTimeField(blank=True, null=True)),
                ("reason", models.TextField(blank=True, null=True)),
                (
                    "approved",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("approved", "Approved"),
                            ("pending", "Pending"),
                            ("rejected", "Rejected"),
                        ],
                        default="",
                        max_length=16,
                        verbose_name="Pending",
                    ),
                ),
                (
                    "payment_user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="payment_user",
                        to="taxsystem.paymentsystem",
                    ),
                ),
            ],
            options={
                "verbose_name": "Tax Payments",
                "verbose_name_plural": "Tax Payments",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Members",
            fields=[
                ("character_name", models.CharField(db_index=True, max_length=100)),
                (
                    "character_id",
                    models.PositiveIntegerField(primary_key=True, serialize=False),
                ),
                (
                    "status",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("active", "Active"),
                            ("missing", "Missing"),
                            ("noaccount", "Unregistered"),
                            ("is_alt", "Is Alt"),
                        ],
                        default="active",
                        max_length=10,
                        verbose_name="Status",
                    ),
                ),
                ("notice", models.TextField(blank=True, null=True)),
                (
                    "corporation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="owner",
                        to="taxsystem.owneraudit",
                    ),
                ),
            ],
            options={
                "verbose_name": "Tax Member System",
                "verbose_name_plural": "Tax Member Systems",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CorporationWalletDivision",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(default=None, max_length=100, null=True)),
                ("balance", models.DecimalField(decimal_places=2, max_digits=20)),
                ("division", models.IntegerField()),
                (
                    "corporation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="corporation_division",
                        to="taxsystem.owneraudit",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CorporationWalletJournalEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=20, null=True
                    ),
                ),
                (
                    "balance",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=20, null=True
                    ),
                ),
                ("context_id", models.BigIntegerField(default=None, null=True)),
                (
                    "context_id_type",
                    models.CharField(
                        choices=[
                            ("structure_id", "Structure Id"),
                            ("station_id", "Station Id"),
                            ("market_transaction_id", "Market Transaction Id"),
                            ("character_id", "Character Id"),
                            ("corporation_id", "Corporation Id"),
                            ("alliance_id", "Alliance Id"),
                            ("eve_system", "Eve System"),
                            ("industry_job_id", "Industry Job Id"),
                            ("contract_id", "Contract Id"),
                            ("planet_id", "Planet Id"),
                            ("system_id", "System Id"),
                            ("type_id", "Type Id"),
                        ],
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                ("date", models.DateTimeField()),
                ("description", models.CharField(max_length=500)),
                ("entry_id", models.BigIntegerField()),
                ("reason", models.CharField(default=None, max_length=500, null=True)),
                ("ref_type", models.CharField(max_length=72)),
                (
                    "tax",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=20, null=True
                    ),
                ),
                ("tax_receiver_id", models.IntegerField(default=None, null=True)),
                (
                    "division",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="taxsystem.corporationwalletdivision",
                    ),
                ),
                (
                    "first_party",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "second_party",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "default_permissions": (),
                "indexes": [
                    models.Index(fields=["date"], name="taxsystem_c_date_1a3f46_idx"),
                    models.Index(
                        fields=["amount"], name="taxsystem_c_amount_3f4bd8_idx"
                    ),
                    models.Index(
                        fields=["entry_id"], name="taxsystem_c_entry_i_bd9090_idx"
                    ),
                    models.Index(
                        fields=["ref_type"], name="taxsystem_c_ref_typ_5f82e6_idx"
                    ),
                    models.Index(
                        fields=["first_party"], name="taxsystem_c_first_p_d83db8_idx"
                    ),
                    models.Index(
                        fields=["second_party"], name="taxsystem_c_second__2a759e_idx"
                    ),
                ],
            },
        ),
    ]



"""App Tasks"""

import datetime

from celery import shared_task

from django.utils import timezone

from allianceauth.services.tasks import QueueOnce

from taxsystem import app_settings
from taxsystem.decorators import when_esi_is_available
from taxsystem.hooks import get_extension_logger
from taxsystem.models.tax import OwnerAudit
from taxsystem.task_helpers.general_helpers import enqueue_next_task, no_fail_chain
from taxsystem.task_helpers.payment_helpers import (
    update_corporation_payments,
    update_corporation_payments_filter,
)
from taxsystem.task_helpers.tax_helpers import update_corporation_members
from taxsystem.task_helpers.wallet_helpers import update_corporation_wallet_division

logger = get_extension_logger(__name__)


@shared_task
@when_esi_is_available
def update_all_taxsytem(runs: int = 0):
    corps = OwnerAudit.objects.select_related("corporation").all()
    for corp in corps:
        update_corp.apply_async(args=[corp.corporation.corporation_id])
        runs = runs + 1


@shared_task(bind=True, base=QueueOnce)
def update_corp(self, corp_id, force_refresh=False):  # pylint: disable=unused-argument
    class SkipDates:
        """Skip Dates for Updates"""

        wallet = timezone.now() - datetime.timedelta(
            hours=app_settings.TAXSYSTEM_CORP_WALLET_SKIP_DATE
        )
        members = timezone.now() - datetime.timedelta(
            days=app_settings.TAXSYSTEM_CORP_MEMBERS_SKIP_DATE
        )
        payments = timezone.now() - datetime.timedelta(
            days=app_settings.TAXSYSTEM_CORP_PAYMENTS_SKIP_DATE
        )
        payment_system = timezone.now() - datetime.timedelta(
            days=app_settings.TAXSYSTEM_CORP_PAYMENT_SYSTEM_SKIP_DATE
        )

    corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)
    logger.debug("Processing Audit Updates for %s", corp.corporation.corporation_name)

    que = []
    mindt = timezone.now() - datetime.timedelta(days=90)

    if (corp.last_update_wallet or mindt) <= SkipDates.wallet or force_refresh:
        que.append(update_corp_wallet.si(corp_id, force_refresh=force_refresh))
    if (corp.last_update_members or mindt) <= SkipDates.members or force_refresh:
        que.append(update_corp_members.si(corp_id, force_refresh=force_refresh))
    if (corp.last_update_payments or mindt) <= SkipDates.payments or force_refresh:
        que.append(update_corp_payments.si(corp_id))
    if (
        corp.last_update_payment_system or mindt
    ) <= SkipDates.payment_system or force_refresh:
        que.append(update_corp_payments_filter.si(corp_id))

    enqueue_next_task(que)

    logger.debug("Queued Updates for %s", corp.corporation.corporation_name)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["corp_id"]},
    name="taxsystem.tasks.update_corp_wallet",
)
@no_fail_chain
def update_corp_wallet(
    self, corp_id, force_refresh=False, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_corporation_wallet_division(corp_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["corp_id"]},
    name="taxsystem.tasks.update_corp_members",
)
@no_fail_chain
def update_corp_members(
    self, corp_id, force_refresh=False, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_corporation_members(corp_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["corp_id"]},
    name="taxsystem.tasks.update_corp_payments",
)
@no_fail_chain
def update_corp_payments(
    self, corp_id, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_corporation_payments(corp_id)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["corp_id"]},
    name="taxsystem.tasks.update_corp_payments_filter",
)
@no_fail_chain
def update_corp_payments_filter(
    self, corp_id, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_corporation_payments_filter(corp_id)



from ninja import NinjaAPI

from django.contrib.humanize.templatetags.humanize import intcomma
from django.template.loader import render_to_string
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from taxsystem.api.helpers import get_corporation
from taxsystem.helpers import lazy
from taxsystem.hooks import get_extension_logger
from taxsystem.models.tax import Members, Payments, PaymentSystem

logger = get_extension_logger(__name__)


class CorporationApiEndpoints:
    tags = ["Corporation Tax System"]

    # pylint: disable=too-many-statements
    def __init__(self, api: NinjaAPI):
        @api.get(
            "corporation/{corporation_id}/view/members/",
            response={200: list, 403: str, 404: str},
            tags=self.tags,
        )
        def get_members(request, corporation_id: int):
            perms, corp = get_corporation(request, corporation_id)

            if perms is False:
                return 403, "Permission Denied"

            if corp is None:
                return 404, "Corporation Not Found"

            corporation_dict = {}

            members = Members.objects.filter(corporation=corp)

            for member in members:
                corporation_dict[member.character_id] = {
                    "character_id": member.character_id,
                    "character_portrait": lazy.get_character_portrait_url(
                        member.character_id, size=32, as_html=True
                    ),
                    "character_name": member.character_name,
                    "is_faulty": member.is_faulty,
                    "status": member.get_status_display(),
                    "joined": lazy.str_normalize_time(member.joined, hours=False),
                    "actions": "",
                }

            output = []
            output.append({"corporation": corporation_dict})

            return output

        @api.get(
            "corporation/{corporation_id}/view/paymentsystem/",
            response={200: list, 403: str, 404: str},
            tags=self.tags,
        )
        # pylint: disable=too-many-locals
        def get_paymentsystem(request, corporation_id: int):
            perms, corp = get_corporation(request, corporation_id)

            if perms is False:
                return 403, "Permission Denied"

            if corp is None:
                return 404, "Corporation Not Found"

            payment_system = PaymentSystem.objects.filter(
                corporation=corp
            ).select_related("user", "user__user", "user__main_character")

            payment_dict = {}

            for user in payment_system:
                actions = ""
                if perms is True:
                    url = "taxsystem/forms/paymentsystem/switchuser.html"
                    if user.is_active:
                        confirm_text = ""
                        confirm_text += _("Are you sure to Confirm")
                        confirm_text += f"?<br><span class='fw-bold'>Deactivate {user.name} (ID: {user.pk}) "
                        title = _("Deactivate User")
                        settings = {
                            "icon": "fas fa-eye-low-vision",
                            "color": "warning",
                            "title": title,
                        }
                    else:
                        confirm_text = ""
                        confirm_text += _("Are you sure to Confirm")
                        confirm_text += f"?<br><span class='fw-bold'>Activate {user.name} (ID: {user.pk}) "
                        title = _("Activate User")
                        settings = {
                            "icon": "fas fa-eye",
                            "color": "success",
                            "title": title,
                        }

                    actions = format_html(
                        "<td>{}</td>",
                        render_to_string(
                            template_name=url,
                            context={
                                "corporation_id": corporation_id,
                                "payment_system": user,
                                "title": title,
                                "confirm_text": confirm_text,
                                "settings": settings,
                            },
                            request=request,
                        ),
                    )
                has_paid_filter = _("Yes") if user.has_paid else _("No")
                has_paid = {
                    "display": lazy.get_bool_icon_html(value=user.has_paid),
                    "sort": has_paid_filter,
                }
                character_id = user.user.main_character.character_id
                payment_dict[character_id] = {
                    "character_id": character_id,
                    "character_portrait": lazy.get_character_portrait_url(
                        character_id=character_id,
                        size=32,
                        as_html=True,
                    ),
                    "character_name": user.user.main_character.character_name,
                    "alts": user.get_alt_ids(),
                    "status": user.get_payment_status(),
                    "wallet": user.payment_pool,
                    "has_paid": has_paid,
                    "has_paid_filter": has_paid_filter,
                    "has_paid_raw": user.has_paid,
                    "last_paid": lazy.str_normalize_time(user.last_paid, hours=True),
                    "is_active": user.is_active,
                    "actions": actions,
                }

            output = []
            output.append({"corporation": payment_dict})

            return output

        @api.get(
            "corporation/{corporation_id}/view/payments/",
            response={200: list, 403: str, 404: str},
            tags=self.tags,
        )
        def get_payments(request, corporation_id: int):
            perms, corp = get_corporation(request, corporation_id)

            if corp is None:
                return 404, "Corporation Not Found"

            payment_system = Payments.objects.filter(payment_user__corporation=corp)

            payments_dict = {}

            for payment in payment_system:
                try:
                    character_id = payment.payment_user.user.main_character.character_id
                except AttributeError:
                    character_id = 0

                actions = ""
                if perms is True:
                    if payment.payment_status in [
                        Payments.States.PENDING,
                        Payments.States.NEEDS_APPROVAL,
                    ]:
                        amount = intcomma(payment.amount)
                        confirm_text = ""
                        confirm_text += _("Are you sure to Confirm")
                        confirm_text += f"?<br><span class='fw-bold'>{amount} ISK (ID: {payment.pk}) "
                        confirm_text += _("from")
                        confirm_text += f" {payment.payment_user.name}</span>"

                        actions = format_html(
                            "<td>{}</td>",
                            render_to_string(
                                "taxsystem/forms/approve.html",
                                {
                                    "corporation_id": corporation_id,
                                    "payment": payment,
                                    "title": _("Approve Payment"),
                                    "confirm_text": confirm_text,
                                    "request": request,
                                },
                                request=request,
                            ),
                        )

                payments_dict[payment.pk] = {
                    "payment_id": payment.pk,
                    "date": payment.date,
                    "character_portrait": lazy.get_character_portrait_url(
                        character_id, size=32, as_html=True
                    ),
                    "character_name": payment.payment_user.name,
                    "amount": payment.amount,
                    "status": payment.get_payment_status_display(),
                    "approved": payment.get_approved_display(),
                    "system": payment.get_system_display(),
                    "payment_date": payment.formatted_payment_date(),
                    "actions": actions,
                }

            output = []
            output.append({"corporation": payments_dict})

            return output

        @api.get(
            "corporation/{corporation_id}/view/dashboard/",
            response={200: dict, 403: str, 404: str},
            tags=self.tags,
        )
        def get_dashboard(request, corporation_id: int):
            perms, corp = get_corporation(request, corporation_id)

            if perms is False:
                return 403, "Permission Denied"

            if corp is None:
                return 404, "Corporation Not Found"

            corporation_name = corp.name
            corporation_id = corp.corporation.corporation_id
            corporation_logo = lazy.get_corporation_logo_url(
                corporation_id, size=64, as_html=True
            )
            last_update_wallet = lazy.str_normalize_time(
                corp.last_update_wallet, hours=True
            )
            last_update_members = lazy.str_normalize_time(
                corp.last_update_members, hours=True
            )
            last_update_payments = lazy.str_normalize_time(
                corp.last_update_payments, hours=True
            )
            last_update_payment_system = lazy.str_normalize_time(
                corp.last_update_payment_system, hours=True
            )
            corporation_tax_amount = corp.tax_amount
            corporation_tax_period = corp.tax_period

            output = {
                "corporation_name": corporation_name,
                "corporation_id": corporation_id,
                "corporation_logo": corporation_logo,
                "last_update_wallet": last_update_wallet,
                "last_update_members": last_update_members,
                "last_update_payments": last_update_payments,
                "last_update_payment_system": last_update_payment_system,
                "tax_amount": corporation_tax_amount,
                "tax_period": corporation_tax_period,
            }

            return output



"""PvE Views"""

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.exceptions import ValidationError
from django.core.handlers.wsgi import WSGIRequest
from django.db import IntegrityError, transaction
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from esi.decorators import token_required

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from taxsystem.api.helpers import get_corporation
from taxsystem.helpers.views import add_info_to_context
from taxsystem.models import OwnerAudit
from taxsystem.models.tax import Payments, PaymentSystem
from taxsystem.tasks import update_corp

from .hooks import get_extension_logger

logger = get_extension_logger(__name__)


@login_required
@permission_required("taxsystem.basic_access")
def index(request):
    context = {
        "title": _("Tax System"),
    }
    context = add_info_to_context(request, context)
    return render(request, "taxsystem/index.html", context=context)


@login_required
@permission_required("taxsystem.basic_access")
def administration(request, corporation_pk):
    """Manage View"""

    context = {
        "entity_pk": corporation_pk,
        "entity_type": "corporation",
        "title": _("Administration"),
    }
    context = add_info_to_context(request, context)

    return render(request, "taxsystem/view/manage.html", context=context)


@login_required
@permission_required("taxsystem.basic_access")
def payments(request, corporation_pk):
    """Payments View"""

    if corporation_pk == 0:
        try:
            corporation_pk = request.user.profile.main_character.corporation_id
        except AttributeError:
            messages.error(request.user, "No Main Character found")

    context = {
        "entity_pk": corporation_pk,
        "entity_type": "corporation",
        "title": _("Payments"),
    }
    context = add_info_to_context(request, context)

    return render(request, "taxsystem/view/payments.html", context=context)


@login_required
@permission_required("taxsystem.create_access")
@token_required(scopes=OwnerAudit.get_esi_scopes())
def add_corp(request, token):
    char = get_object_or_404(EveCharacter, character_id=token.character_id)
    corp, _ = EveCorporationInfo.objects.get_or_create(
        corporation_id=char.corporation_id,
        defaults={
            "member_count": 0,
            "corporation_ticker": char.corporation_ticker,
            "corporation_name": char.corporation_name,
        },
    )

    OwnerAudit.objects.update_or_create(
        corporation=corp,
        defaults={
            "name": corp.corporation_name,
            "active": True,
        },
    )

    update_corp.apply_async(
        args=[char.corporation_id], kwargs={"force_refresh": True}, priority=6
    )
    msg = _("{corporation_name} successfully added/updated to Tax System").format(
        corporation_name=corp.corporation_name,
    )
    messages.info(request, msg)
    return redirect("taxsystem:index")


@login_required
@permission_required("taxsystem.manage_access")
def overview(request):
    """Overview of the tax system"""

    context = {}
    context = add_info_to_context(request, context)

    return render(request, "taxsystem/admin/overview.html", context=context)


@login_required
@permission_required("taxsystem.manage_access")
@require_POST
def approve_payment(request, corporation_id: int, payment_pk: int):
    # Check Permission
    perms, corp = get_corporation(request, corporation_id)
    previous_url = request.headers.get("referer", "taxsystem:payments")

    if not perms:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return redirect(previous_url)

    try:
        with transaction.atomic():
            payment = Payments.objects.get(
                payment_user__corporation=corp, pk=payment_pk
            )
            if payment.payment_status in [
                Payments.States.PENDING,
                Payments.States.NEEDS_APPROVAL,
            ]:
                payment.approved = Payments.Approval.APPROVED
                payment.payment_status = Payments.States.PAID
                payment.system = Payments.Systems.MANUAL
                payment.save()

                payment_user = PaymentSystem.objects.get(
                    corporation=corp, user=payment.payment_user.user
                )
                payment_user.payment_pool += payment.amount
                payment_user.save()

                msg = _("Payment ID: %s successfully approved") % payment.pk
            else:
                msg = _("Payment ID: %s is already edited") % payment.pk
    except IntegrityError:
        msg = _("Transaction failed. Please try again.")

    messages.info(request, msg)
    return redirect(previous_url)


@login_required
@permission_required("taxsystem.manage_access")
@require_POST
def switch_user(request, corporation_id: int, user_pk: int):
    # Check Permission
    perms, corp = get_corporation(request, corporation_id)
    previous_url = request.headers.get("referer", "taxsystem:payments")

    if not perms:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return redirect(previous_url)

    try:
        with transaction.atomic():
            payment_system = PaymentSystem.objects.get(corporation=corp, pk=user_pk)
            if payment_system.is_active:
                payment_system.status = PaymentSystem.States.DEACTIVATED
                msg = (
                    _("Payment System User: %s successfully deactivated")
                    % payment_system.name
                )
            else:
                payment_system.status = PaymentSystem.States.ACTIVE
                msg = (
                    _("Payment System User: %s successfully activated")
                    % payment_system.name
                )
            payment_system.save()
    except IntegrityError:
        msg = _("Transaction failed. Please try again.")

    messages.info(request, msg)
    return redirect(previous_url)


@csrf_exempt
def update_tax_amount(request: WSGIRequest, corporation_id: int):
    if request.method == "POST":
        value = float(request.POST.get("value"))

        try:
            if value < 0:
                return JsonResponse(
                    {"message": _("Please enter a valid number")}, status=400
                )
        except ValueError:
            return JsonResponse(
                {"message": _("Please enter a valid number")}, status=400
            )

        # Check Permission
        perms, corp = get_corporation(request, corporation_id)

        if not perms:
            return JsonResponse({"message": _("Permission Denied")}, status=403)

        try:
            corp.tax_amount = value
            corp.save()
        except ValidationError:
            return JsonResponse(
                {"message": _("Please enter a valid number")}, status=400
            )
        return JsonResponse({"message": ""}, status=200)
    return JsonResponse({"message": _("Invalid request method")}, status=405)


@csrf_exempt
def update_tax_period(request: WSGIRequest, corporation_id: int):
    if request.method == "POST":
        value = float(request.POST.get("value"))

        try:
            if value < 0:
                return JsonResponse(
                    {"message": _("Please enter a valid number")}, status=400
                )
        except ValueError:
            return JsonResponse(
                {"message": _("Please enter a valid number")}, status=400
            )

        # Check Permission
        perms, corp = get_corporation(request, corporation_id)

        if not perms:
            return JsonResponse({"message": _("Permission Denied")}, status=403)

        try:
            corp.tax_period = value
            corp.save()
        except ValidationError:
            return JsonResponse(
                {"message": _("Please enter a valid number")}, status=400
            )
        return JsonResponse({"message": ""}, status=200)
    return JsonResponse({"message": _("Invalid request method")}, status=405)



"""Tax Helpers"""

from django.utils import timezone
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import UserProfile

from taxsystem.hooks import get_extension_logger
from taxsystem.models.tax import (
    Members,
    OwnerAudit,
    PaymentSystem,
)
from taxsystem.providers import esi
from taxsystem.task_helpers.etag_helpers import NotModifiedError, etag_results
from taxsystem.task_helpers.general_helpers import get_corp_token

logger = get_extension_logger(__name__)


# pylint: disable=too-many-locals
def update_corporation_members(corp_id, force_refresh=False):
    """Update corporation members"""
    audit_corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)

    logger.debug(
        "Updating members for: %s",
        audit_corp.corporation.corporation_name,
    )

    req_scopes = [
        "esi-corporations.read_corporation_membership.v1",
        "esi-corporations.track_members.v1",
    ]

    req_roles = ["CEO", "Director"]

    token = get_corp_token(corp_id, req_scopes, req_roles)

    # pylint: disable=duplicate-code
    if not token:
        logger.debug("No valid token for: %s", audit_corp.corporation.corporation_name)
        return "No Tokens"

    try:
        _current_members_ids = set(
            Members.objects.filter(corporation=audit_corp).values_list(
                "character_id", flat=True
            )
        )
        members_ob = (
            esi.client.Corporation.get_corporations_corporation_id_membertracking(
                corporation_id=audit_corp.corporation.corporation_id,
            )
        )

        members = etag_results(members_ob, token, force_refresh=force_refresh)

        _esi_members_ids = [member.get("character_id") for member in members]
        _old_members = []
        _new_members = []

        characters = EveEntity.objects.bulk_resolve_names(_esi_members_ids)

        for member in members:
            character_id = member.get("character_id")
            joined = member.get("start_date")
            logon_date = member.get("logon_date")
            logged_off = member.get("logoff_date")
            character_name = characters.to_name(character_id)
            member_item = Members(
                corporation=audit_corp,
                character_id=character_id,
                character_name=character_name,
                joined=joined,
                logon=logon_date,
                logged_off=logged_off,
                status=Members.States.ACTIVE,
            )
            if character_id in _current_members_ids:
                _old_members.append(member_item)
            else:
                _new_members.append(member_item)

        # Set missing members
        old_member_ids = {member.character_id for member in _old_members}
        missing_members_ids = _current_members_ids - old_member_ids

        if missing_members_ids:
            Members.objects.filter(
                corporation=audit_corp, character_id__in=missing_members_ids
            ).update(status=Members.States.MISSING)
            logger.debug(
                "Marked %s missing members for: %s",
                len(missing_members_ids),
                audit_corp.corporation.corporation_name,
            )
        if _old_members:
            Members.objects.bulk_update(
                _old_members,
                ["character_name", "status", "logon", "logged_off"],
            )
            logger.debug(
                "Updated %s members for: %s",
                len(_old_members),
                audit_corp.corporation.corporation_name,
            )
        if _new_members:
            Members.objects.bulk_create(_new_members, ignore_conflicts=True)
            logger.debug(
                "Added %s new members for: %s",
                len(_new_members),
                audit_corp.corporation.corporation_name,
            )

        # Update payment users
        update_payment_users(corp_id, _esi_members_ids)

        audit_corp.last_update_members = timezone.now()
        audit_corp.save()

        logger.debug(
            "Corp %s - Old Members: %s, New Members: %s, Missing: %s",
            audit_corp.name,
            len(_old_members),
            len(_new_members),
            len(missing_members_ids),
        )
    except NotModifiedError:
        logger.debug(
            "No changes detected for: %s", audit_corp.corporation.corporation_name
        )
    return ("Finished Members for %s", audit_corp.corporation.corporation_name)


# pylint: disable=too-many-branches
def update_payment_users(corp_id, members_ids):
    """Update payment users for a corporation."""
    audit_corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)

    logger.debug(
        "Updating payment system for: %s",
        audit_corp.corporation.corporation_name,
    )

    accounts = UserProfile.objects.filter(
        main_character__isnull=False,
        main_character__corporation_id=audit_corp.corporation.corporation_id,
    ).select_related(
        "user__profile__main_character",
        "main_character__character_ownership",
        "main_character__character_ownership__user__profile",
        "main_character__character_ownership__user__profile__main_character",
    )

    members = Members.objects.filter(corporation=audit_corp)

    if not accounts:
        logger.debug(
            "No valid accounts for: %s", audit_corp.corporation.corporation_name
        )
        return "No Accounts"

    items = []
    for account in accounts:
        alts = account.user.character_ownerships.all().values_list(
            "character__character_id", flat=True
        )
        main = account.main_character

        if any(alt in members_ids for alt in alts):
            # Remove alts from list
            for alt in alts:
                if alt in members_ids:
                    members_ids.remove(alt)
                    if alt != main.character_id:
                        # Update the status of the member to alt
                        members.filter(character_id=alt).update(
                            status=Members.States.IS_ALT
                        )
            try:
                existing_payment_system = PaymentSystem.objects.get(user=account)
                if existing_payment_system.status != PaymentSystem.States.DEACTIVATED:
                    existing_payment_system.status = PaymentSystem.States.ACTIVE
                    existing_payment_system.save()
            except PaymentSystem.DoesNotExist:
                items.append(
                    PaymentSystem(
                        name=main.character_name,
                        corporation=audit_corp,
                        user=account,
                        status=PaymentSystem.States.ACTIVE,
                    )
                )
        else:
            # Set the account to inactive if none of the conditions are met
            try:
                existing_payment_system = PaymentSystem.objects.get(user=account)
                existing_payment_system.status = PaymentSystem.States.INACTIVE
                existing_payment_system.save()
            except PaymentSystem.DoesNotExist:
                pass

    if members_ids:
        # Mark members without accounts
        for member_id in members_ids:
            members.filter(character_id=member_id).update(
                status=Members.States.NOACCOUNT
            )

        logger.debug(
            "Marked %s members without accounts for: %s",
            len(members_ids),
            audit_corp.corporation.corporation_name,
        )

    if items:
        PaymentSystem.objects.bulk_create(items, ignore_conflicts=True)
        logger.debug(
            "Added %s new payment users for: %s",
            len(items),
            audit_corp.corporation.corporation_name,
        )
    else:
        logger.debug(
            "No new payment user for: %s",
            audit_corp.corporation.corporation_name,
        )

    return ("Finished payment system for %s", audit_corp.corporation.corporation_name)



"""App URLs"""

from django.urls import path, re_path

from taxsystem import views
from taxsystem.api import api

app_name: str = "taxsystem"

urlpatterns = [
    path("", views.index, name="index"),
    # -- API System
    re_path(r"^api/", api.urls),
    # -- Tax System
    path("corporation/overview/", views.overview, name="overview"),
    path(
        "corporation/<int:corporation_pk>/view/administration/",
        views.administration,
        name="administration",
    ),
    path(
        "corporation/<int:corporation_pk>/view/payments/",
        views.payments,
        name="payments",
    ),
    # --- Tax Administration
    # -- Tax Payments
    path("corporation/add/", views.add_corp, name="add_corp"),
    path(
        "corporation/<int:corporation_id>/payment/<int:payment_pk>/approve/",
        views.approve_payment,
        name="approve_payment",
    ),
    # -- Tax Manage
    path(
        "corporation/<int:corporation_id>/manage/update_tax/",
        views.update_tax_amount,
        name="update_tax_amount",
    ),
    path(
        "corporation/<int:corporation_id>/manage/update_period/",
        views.update_tax_period,
        name="update_tax_period",
    ),
    # -- Tax Payment System
    path(
        "corporation/<int:corporation_id>/manage/user/<int:user_pk>/switch_user/",
        views.switch_user,
        name="switch_user",
    ),
]



"""
Decorators
"""

import time
from functools import wraps

from app_utils.esi import EsiDailyDowntime, fetch_esi_status

from taxsystem.app_settings import IS_TESTING
from taxsystem.hooks import get_extension_logger

logger = get_extension_logger(__name__)


def when_esi_is_available(func):
    """Make sure the decorated task only runs when esi is available.

    Raise exception when ESI is offline.
    Complete the task without running it when downtime is detected.

    Automatically disabled during tests.
    """

    @wraps(func)
    def outer(*args, **kwargs):
        if IS_TESTING is not True:
            try:
                fetch_esi_status().raise_for_status()
            except EsiDailyDowntime:
                logger.info("Daily Downtime detected. Aborting.")
                return None  # function will not run

        return func(*args, **kwargs)

    return outer


def log_timing(logs):
    """
    A Decirator to log the time a function takes to run.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result = func(*args, **kwargs)
            end_time = time.time()
            logs.debug(
                "TIME: %s run for %s seconds with args: %s",
                end_time - start_time,
                func.__name__,
                args,
            )
            return result

        return wrapper

    return decorator



"""Shared ESI client for Voices of War."""

from esi.clients import EsiClientProvider

from taxsystem import __title__, __version__

esi = EsiClientProvider(app_info_text=f"{__title__} v{__version__}")



from taxsystem import models


def get_corporation(
    request, corporation_id
) -> tuple[bool | None, models.tax.OwnerAudit | None]:
    """Get Corporation and check permissions"""
    perms = True

    try:
        corp = models.OwnerAudit.objects.get(corporation__corporation_id=corporation_id)
    except models.OwnerAudit.DoesNotExist:
        return None, None

    # Check access
    visible = models.OwnerAudit.objects.visible_to(request.user)
    if corp not in visible:
        perms = False
    return perms, corp



"""Admin models"""

from django.contrib import admin
from django.utils.html import format_html

from allianceauth.eveonline.evelinks import eveimageserver

from taxsystem.models.filters import FilterAmount, FilterReason, SmartFilter, SmartGroup
from taxsystem.models.tax import OwnerAudit


@admin.register(OwnerAudit)
class OwnerAuditAdmin(admin.ModelAdmin):
    list_display = (
        "_entity_pic",
        "_corporation__corporation_id",
        "_corporation__corporation_name",
        "_last_update_wallet",
    )

    list_display_links = (
        "_entity_pic",
        "_corporation__corporation_id",
        "_corporation__corporation_name",
    )

    list_select_related = ("corporation",)

    ordering = ["corporation__corporation_name"]

    search_fields = ["corporation__corporation_name", "corporation__corporation_id"]

    actions = [
        "delete_objects",
    ]

    @admin.display(description="")
    def _entity_pic(self, obj: OwnerAudit):
        eve_id = obj.corporation.corporation_id
        return format_html(
            '<img src="{}" class="img-circle">',
            eveimageserver._eve_entity_image_url("corporation", eve_id, 32),
        )

    @admin.display(description="Corporation ID", ordering="corporation__corporation_id")
    def _corporation__corporation_id(self, obj: OwnerAudit):
        return obj.corporation.corporation_id

    @admin.display(
        description="Corporation Name", ordering="corporation__corporation_name"
    )
    def _corporation__corporation_name(self, obj: OwnerAudit):
        return obj.corporation.corporation_name

    @admin.display(description="Last Update Wallet", ordering="last_update_wallet")
    def _last_update_wallet(self, obj: OwnerAudit):
        return obj.last_update_wallet

    @admin.display(description="Last Update Members", ordering="last_update_members")
    def _last_update_members(self, obj: OwnerAudit):
        return obj.last_update_members

    @admin.display(description="Last Update Payments", ordering="last_update_payments")
    def _last_update_payments(self, obj: OwnerAudit):
        return obj.last_update_payments

    # pylint: disable=unused-argument
    def has_add_permission(self, request):
        return False

    # pylint: disable=unused-argument
    def has_change_permission(self, request, obj=None):
        return False


@admin.register(FilterAmount)
class FilterAmountAdmin(admin.ModelAdmin):
    list_display = ("name", "description", "amount")


@admin.register(FilterReason)
class FilterReasonAdmin(admin.ModelAdmin):
    list_display = ("name", "description", "reason")


@admin.register(SmartFilter)
class SmartfilterAdmin(admin.ModelAdmin):
    # pylint: disable=unused-argument
    def has_add_permission(self, request, obj=None):
        return False

    list_display = ["__str__"]


@admin.register(SmartGroup)
class SmartGroupAdmin(admin.ModelAdmin):
    filter_horizontal = ["filters"]
    list_display = [
        "__str__",
        "enabled",
        "display_filters",
        "last_update",
        "corporation",
    ]



"""General Model"""

from django.db import models
from django.utils.translation import gettext_lazy as _


class General(models.Model):
    """General model for app permissions"""

    class Meta:
        managed = False
        permissions = (
            ("basic_access", _("Can access this app")),
            ("manage_access", _("Can manage Tax System")),
            ("create_access", _("Can add Corporation")),
        )
        default_permissions = ()



# Django
from django.db import models

# AA Voices of War
from taxsystem.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class WalletQuerySet(models.QuerySet):
    pass


class WalletManagerBase(models.Manager):
    pass


WalletManager = WalletManagerBase.from_queryset(WalletQuerySet)



"""
Core Helpers
"""

from functools import wraps

# pylint: disable=no-name-in-module
from celery import signature
from celery_once import AlreadyQueued

from esi.errors import TokenError
from esi.models import Token

from allianceauth.eveonline.models import EveCharacter

from taxsystem.hooks import get_extension_logger
from taxsystem.providers import esi

logger = get_extension_logger(__name__)


def get_token(character_id: int, scopes: list) -> Token:
    """
    Helper method to get a valid token for a specific character with specific scopes.

    Parameters
    ----------
    character_id: `int`
    scopes: `int`

    Returns
    ----------
    `class`: esi.models.Token or False

    """
    token = (
        Token.objects.filter(character_id=character_id)
        .require_scopes(scopes)
        .require_valid()
        .first()
    )
    if token:
        return token
    return False


def get_corp_token(corp_id, scopes, req_roles):
    """
    Helper method to get a token for a specific character from a specific corp with specific scopes

    Parameters
    ----------
    corp_id: `int`
    scopes: `int`
    req_roles: `list`

    Returns
    ----------
    `class`: esi.models.Token or False

    """
    if "esi-characters.read_corporation_roles.v1" not in scopes:
        scopes.append("esi-characters.read_corporation_roles.v1")

    char_ids = EveCharacter.objects.filter(corporation_id=corp_id).values(
        "character_id"
    )
    tokens = Token.objects.filter(character_id__in=char_ids).require_scopes(scopes)

    for token in tokens:
        try:
            roles = esi.client.Character.get_characters_character_id_roles(
                character_id=token.character_id, token=token.valid_access_token()
            ).result()

            has_roles = False
            for role in roles.get("roles", []):
                if role in req_roles:
                    has_roles = True

            if has_roles:
                return token
        except TokenError as e:
            logger.error(
                "Token ID: %s (%s)",
                token.pk,
                e,
            )
    return False


def enqueue_next_task(chain):
    """
    Queue next task, and attach the rest of the chain to it.
    """
    while len(chain):
        _t = chain.pop(0)
        _t = signature(_t)
        _t.kwargs.update({"chain": chain})
        try:
            _t.apply_async(priority=9)
        except AlreadyQueued:
            # skip this task as it is already in the queue
            logger.debug("Skipping task as its already queued %s", _t)
            continue
        break


def no_fail_chain(func):
    """
    Decorator to chain tasks provided in the chain kwargs regardless of task failures.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        excp = None
        _ret = None
        try:
            _ret = func(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-exception-caught
            excp = e
        finally:
            _chn = kwargs.get("chain", [])
            enqueue_next_task(_chn)
            if excp:
                raise excp
        return _ret

    return wrapper



# Generated by Django 4.2.11 on 2025-02-09 21:00

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("taxsystem", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="FilterDate",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("description", models.CharField(max_length=255)),
                ("date", models.DateTimeField()),
            ],
            options={
                "verbose_name": "Filter Date",
                "verbose_name_plural": "Filter Dates",
            },
        ),
        migrations.RenameField(
            model_name="owneraudit",
            old_name="last_update_filters",
            new_name="last_update_payment_system",
        ),
        migrations.AddField(
            model_name="members",
            name="joined",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="members",
            name="logged_off",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="members",
            name="logon",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="owneraudit",
            name="tax_amount",
            field=models.DecimalField(
                decimal_places=0,
                default=0,
                help_text="Tax Amount in ISK that is set for the corporation. Max 16 Digits",
                max_digits=16,
                validators=[django.core.validators.MaxValueValidator(9999999999999999)],
            ),
        ),
        migrations.AddField(
            model_name="owneraudit",
            name="tax_period",
            field=models.PositiveIntegerField(
                default=30,
                help_text="Tax Period in days for the corporation. Max 365 days. Default: 30 days",
                validators=[django.core.validators.MaxValueValidator(365)],
            ),
        ),
        migrations.AddField(
            model_name="payments",
            name="notified",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="payments",
            name="system",
            field=models.CharField(
                blank=True,
                choices=[("automatic", "Automatic"), ("manual", "Manual")],
                default="",
                max_length=16,
                verbose_name="System",
            ),
        ),
        migrations.AddField(
            model_name="paymentsystem",
            name="last_paid",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="paymentsystem",
            name="payment_pool",
            field=models.DecimalField(
                decimal_places=0,
                default=0,
                help_text="Payment Pool in ISK that is set for the corporation. Max 16 Digits",
                max_digits=16,
                validators=[
                    django.core.validators.MaxValueValidator(9999999999999999),
                    django.core.validators.MinValueValidator(-9999999999999999),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="payments",
            name="payment_status",
            field=models.CharField(
                blank=True,
                choices=[
                    ("paid", "Paid"),
                    ("pending", "Pending"),
                    ("failed", "Failed"),
                    ("needs_approval", "Needs Approval"),
                ],
                default="",
                max_length=16,
                verbose_name="Status",
            ),
        ),
    ]



"""Models for Filters."""

# Django

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from taxsystem.models.tax import OwnerAudit, Payments


class SmartFilter(models.Model):
    """Model to hold a filter and its settings"""

    class Meta:
        verbose_name = _("Smart Filter Binding")
        verbose_name_plural = _("Smart Filters Catalog")

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="+"
    )
    object_id = models.PositiveIntegerField()
    filter_object = GenericForeignKey("content_type", "object_id")

    def __str__(self):
        try:
            return f"{self.filter_object.name}: {self.filter_object.description}"
        # pylint: disable=broad-exception-caught
        except Exception:
            return f"Error: {self.content_type.app_label}:{self.content_type} {self.object_id} Not Found"


class FilterBase(models.Model):
    """Base Filter Model"""

    name = models.CharField(max_length=100)
    description = models.CharField(max_length=255)

    class Meta:
        abstract = True

    def __str__(self):
        return f"{self.name}: {self.description}"

    def filter(self):
        raise NotImplementedError("Please Create a filter!")

    def filter_containts(self):
        raise NotImplementedError("Please Create a contains filter!")


class FilterAmount(FilterBase):
    """Filter for Amount"""

    class Meta:
        verbose_name = _("Filter Amount")
        verbose_name_plural = _("Filter Amounts")

    amount = models.DecimalField(max_digits=12, decimal_places=0)

    def filter(self):
        return {"amount": self.amount}

    def filter_containts(self):
        return {"amount__gte": self.amount}


class FilterReason(FilterBase):
    """Filter for Reason"""

    class Meta:
        verbose_name = _("Filter Reason")
        verbose_name_plural = _("Filter Reasons")

    reason = models.CharField(max_length=255)

    def filter(self):
        return {"reason": self.reason}

    def filter_containts(self):
        return {"reason__contains": self.reason}


class FilterDate(FilterBase):
    """Filter for Date"""

    class Meta:
        verbose_name = _("Filter Date")
        verbose_name_plural = _("Filter Dates")

    date = models.DateTimeField()

    def filter(self):
        return {"date": self.date}

    def filter_containts(self):
        return {"date__gte": self.date}


class SmartGroup(models.Model):
    """Model to hold a group of filters"""

    corporation = models.OneToOneField(
        OwnerAudit, on_delete=models.CASCADE, related_name="filter_sets"
    )
    description = models.CharField(max_length=255)
    name = models.CharField(max_length=100)
    filters = models.ManyToManyField(SmartFilter)
    last_update = models.DateTimeField(auto_now=True)
    enabled = models.BooleanField(default=True)

    def filter(self, payments: Payments) -> models.QuerySet[Payments]:
        if self.enabled is True:
            q_objects = Q()
            for smart_filter in self.filters.all():
                q_objects &= Q(**smart_filter.filter_object.filter())
            payments = payments.filter(q_objects)
        return payments

    def filter_containts(self, payments: Payments) -> models.QuerySet[Payments]:
        if self.enabled is True:
            q_objects = Q()
            for smart_filter in self.filters.all():
                q_objects |= Q(**smart_filter.filter_object.filter_containts())
            payments = payments.filter(q_objects)
        return payments

    def display_filters(self):
        return ", ".join([str(f) for f in self.filters.all()])

    display_filters.short_description = "Filters"

    def __str__(self):
        return f"{self.name}: {self.description}"
